#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging, io, re
from datetime import datetime
from c8ydm.framework.modulebase import Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage
from c8ydm.core.device_stats import DeviceStats


class MeasurementRequestHandler(Initializer, Listener):
    logger = logging.getLogger(__name__)
    fragment = 'c8y_MeasurementRequestOperation'
    DeviceStats = DeviceStats()
    

    def getMessages(self):
        return []

    def getSupportedOperations(self):
        return ['c8y_MeasurementRequestOperation']
    
    def getSupportedTemplates(self):
        return []

    def _set_executing(self):
        executing = SmartRESTMessage('s/us', '501', [self.fragment])
        #print("Executing MSG send")
        self.agent.publishMessage(executing)

    #datei anhängen
    def _set_success(self):
        success = SmartRESTMessage('s/us', '503', [self.fragment,'CPU:RAM:DISK'])
        #print("Success MSG send")
        self.agent.publishMessage(success)

    def _set_failed(self, reason):
        failed = SmartRESTMessage('s/us', '502', [self.fragment,reason])
        self.logger.error(f'Operation failed, reason: {reason}')
        self.agent.publishMessage(failed)
    
    def _getCPU(self):
        return self.DeviceStats.getCPUStats() 

    def _getDisk(self):
        return self.DeviceStats.getDiskStats() 
    
    def _getMemory(self):
        return self.DeviceStats.getMemoryStats()
    
    def handleOperation(self, message):
        mo_id = self.agent.rest_client.get_internal_id(self.agent.serial)
        try:
            if 's/ds' in message.topic and message.messageId == '517':
                self._set_executing()
                self.logger.info("Sending device stats due to measurement request")
                self.stats = []
                for key,value in self._getCPU().items():
                    self.stats.append(SmartRESTMessage('s/us', '200', ['cpu', key, value]))
                for key,value in self._getDisk().items():
                    self.stats.append(SmartRESTMessage('s/us', '200', ['disk', key, value]))
                for key,value in self._getMemory().items():
                    self.stats.append(SmartRESTMessage('s/us', '200', ['memory', key, value]))
                for i in self.stats:
                    self.agent.publishMessage(i)
                self.logger.debug("Sended device stats due to measurment request")
                self._set_success()
            self.logger.debug("Measurement request handled")
        except Exception as e:
            self._set_failed(e)




"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from c8ydm.client.mqtt_agent import Agent
from c8ydm.client.bootstrap_client import Bootstrap
from c8ydm.client.rest_client import RestClient


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import json
import re
import logging
from operator import contains
import time
import subprocess as sp

from c8ydm.core.apt_package_manager import AptPackageManager
from c8ydm.framework.modulebase import Initializer, Listener
from c8ydm.framework.smartrest import SmartRESTMessage


class SoftwareManager(Listener, Initializer):
    """ Software Update Module"""
    logger = logging.getLogger(__name__)
    apt_package_manager = AptPackageManager()

    def group(self, seq, sep):
        result = [[]]
        for e in seq:
            #logging.info("e: "+str(e) +" sep: " + str(sep))
            if sep not in str(e):
                result[-1].append(e)
            else:
                result[-1].append(e[:e.find(sep)])
                result.append([])

        if result[-1] == []:
            result.pop()  # thx iBug's comment
        return result
    
    def get_filename_from_cd(self, cd):
        """
        Get filename from content-disposition
        """
        if not cd:
            return None
        fname = re.findall('filename=(.+)', cd)
        if len(fname) == 0:
            return None
        return fname[0]

    def handleOperation(self, message):
        try:
            if 's/ds' in message.topic and message.messageId == '528':
                # When multiple operations received just take the first one for further processing
                #self.logger.debug("message received :" + str(message.values))
                messages = self.group(message.values, '\n')[0]
                deviceId = messages.pop(0)
                binary_included = False
                self.logger.info('Software update for device ' +
                                 deviceId + ' with message ' + str(messages))
                executing = SmartRESTMessage(
                    's/us', '501', ['c8y_SoftwareUpdate'])
                self.agent.publishMessage(executing)
                softwareToInstall = [messages[x:x + 4]
                                     for x in range(0, len(messages), 4)]
                for software in softwareToInstall:
                    name = software[0]
                    version = software[1]
                    url = software[2]
                    action = software[3]
                    if 'binaries' in url:
                        # File provided
                        binary_included = True
                if not binary_included:
                    errors = self.apt_package_manager.install_software(
                        softwareToInstall, True)
                    self.logger.info('Finished all software update')
                    if len(errors) == 0:
                        # finished without errors
                        finished = SmartRESTMessage(
                            's/us', '503', ['c8y_SoftwareUpdate'])
                    else:
                        # finished with errors
                        finished = SmartRESTMessage(
                            's/us', '502', ['c8y_SoftwareUpdate', ' - '.join(errors)])
                    self.agent.publishMessage(finished)
                    self.agent.publishMessage(
                        self.apt_package_manager.getInstalledSoftware(False))
                else:
                    # Binary included in software update
                    self.logger.info(f'Software Updated with provided file {url}')
                    file = self.agent.rest_client.download_c8y_binary(url)
                    self.logger.info(f'File to be installed: {file}')
                    if action == 'install' or action == 'update':
                        #result = sp.run(["dpkg","-i", file], stdout=sp.PIPE, stderr=sp.PIPE)
                        result = sp.run(["apt-get","-y","install",file], stdout=sp.PIPE, stderr=sp.PIPE)
                        errors = result.stderr.decode("utf-8")
                        self.logger.info(f'Result of subprocess Error: {errors}')
                        if errors:
                            failed = SmartRESTMessage('s/us', '502', ['c8y_SoftwareUpdate', errors])
                            self.agent.publishMessage(failed)
                        else:
                            finished = SmartRESTMessage(
                                's/us', '503', ['c8y_SoftwareUpdate'])
                            self.agent.publishMessage(finished)
                            self.agent.publishMessage(
                                self.apt_package_manager.getInstalledSoftware(False))
                    
                
            if 's/ds' in message.topic and message.messageId == '516':
                # When multiple operations received just take the first one for further processing
                #self.logger.debug("message received :" + str(message.values))
                messages = self.group(message.values, '\n')[0]
                #self.logger.info("message processed:" + str(messages))
                deviceId = messages.pop(0)
                self.logger.info('Software update for device ' +
                                 deviceId + ' with message ' + str(messages))
                executing = SmartRESTMessage(
                    's/us', '501', ['c8y_SoftwareList'])
                self.agent.publishMessage(executing)
                softwareToInstall = [messages[x:x + 3]
                                     for x in range(0, len(messages), 3)]
                errors = self.apt_package_manager.installSoftware(
                    softwareToInstall, True)
                self.logger.info('Finished all software update')
                if len(errors) == 0:
                    # finished without errors
                    finished = SmartRESTMessage(
                        's/us', '503', ['c8y_SoftwareList'])
                else:
                    # finished with errors
                    finished = SmartRESTMessage(
                        's/us', '502', ['c8y_SoftwareList', ' - '.join(errors)])
                installed_software = self.apt_package_manager.get_installed_software_json(False)
                mo_id = self.agent.rest_client.get_internal_id(self.agent.serial)
                self.agent.rest_client.update_managed_object(mo_id, json.dumps(installed_software))
                self.agent.publishMessage(finished)
                #self.agent.publishMessage(
                #    self.apt_package_manager.getInstalledSoftware(False))
        except Exception as e:
            self.logger.exception(e)
            failed = SmartRESTMessage(
                's/us', '502', ['c8y_SoftwareList', str(e)])
            self.agent.publishMessage(failed)
            failed = SmartRESTMessage(
                's/us', '502', ['c8y_SoftwareUpdate', str(e)])
            self.agent.publishMessage(failed)

    def getSupportedOperations(self):
        return ['c8y_SoftwareUpdate', 'c8y_SoftwareList']

    def getSupportedTemplates(self):
        return []

    def getMessages(self):
        installed_software = self.apt_package_manager.get_installed_software_json(False)
        if self.agent.token_received.wait(timeout=self.agent.refresh_token_interval):
            mo_id = self.agent.rest_client.get_internal_id(self.agent.serial)
            self.agent.rest_client.update_managed_object(mo_id, json.dumps(installed_software))
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
class SmartRESTMessage:

  def __init__(self, topic, messageId, values):
    self.topic = topic
    self.messageId = messageId
    self.values = values

  def getMessage(self):
    values = []
    # Applies the necessary SmartREST escaping to any string passed to the function
    for value in map(str, self.values):
      value = value.replace('"', '""')

      should_escape = '"' in value or ',' in value or '\n' in value or \
        '\r' in value or '\t' in value or value.startswith(' ') or \
          value.endswith(' ')

      if should_escape:
        value = '"{}"'.format(value)

      values.append(value)
    msg = str(self.messageId) + ',' + ','.join(map(str,values))
    return msg.rstrip(', ')



"""  
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging, time, json, time
import subprocess
from c8ydm.framework.modulebase import Listener, Initializer
from c8ydm.framework.smartrest import SmartRESTMessage
import os


class Restart(Listener, Initializer):
    logger = logging.getLogger(__name__)

    def handleOperation(self, message):
        if 's/ds' in message.topic and message.messageId == '510':
            executing = SmartRESTMessage('s/us', '501', ['c8y_Restart'])
            self.agent.publishMessage(executing)
            try:
                if self.agent.simulated:
                    process = subprocess.Popen(["docker","restart",self.serial],stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    process.wait()
                else:
                    os.system('shutdown -r 1')

            except Exception as e:
                failed = SmartRESTMessage('s/us', '502', ['c8y_Restart', 'Error during Restart:' + str(e)])
                self.agent.publishMessage(failed)

    def getSupportedOperations(self):
        return ['c8y_Restart']

    def getSupportedTemplates(self):
        return []

    def getMessages(self):
        response = SmartRESTMessage('s/us', '503', ['c8y_Restart', 'Restart Successful'])
        return [response]


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Copyright (c) 2021 Software AG, Darmstadt, Germany and/or its licensors

SPDX-License-Identifier: Apache-2.0

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
import subprocess
import json
from c8ydm.framework.smartrest import SmartRESTMessage

class DockerWatcher:
    logger = logging.getLogger(__name__)
    docker_active = True

    def get_stats(self):
            try:
                subprocess.Popen(["docker"],stdout=subprocess.PIPE)
                self.docker_active = True
            except Exception as os_ex:
                if self.docker_active:
                    self.logger.warn(f'Docker is not installed, skipping module')
                    self.docker_active = False
                #self.logger.info(f'Docker Status {self.docker_active}')
            if self.docker_active:
                try:
                    rawStats = subprocess.Popen(["docker", "stats", "--no-stream","-a", "--format", "'{{.Container}};{{.Name}};{{.CPUPerc}};{{.MemUsage}}'"],stdout=subprocess.PIPE)
                    list = rawStats.stdout.read().decode('utf-8').split('\n')
                    a = []
                    for i in list:
                        dict = {}
                        for counter,value in enumerate(i.split(';')):
                            if counter == 0:
                                dict["containerID"] = value.replace("'","")
                            if counter == 1:
                                dict["name"] = value
                                nameString = "name=" + str(value)
                                try:
                                    rawStatus = subprocess.Popen(["docker", "ps", "-a", "--format", "'{{.Status}}'", "--filter", nameString ],stdout=subprocess.PIPE)
                                    dict["status"] = rawStatus.stdout.read().decode('utf-8').replace("'","")
                                except:
                                    self.logger.warning('Status from docker for container %s not valid.'% (str(value)))
                                    dict["status"] = "Unknown"
                            if counter == 2:
                                dict["cpu"] = value.replace('%','')
                            if counter == 3:
                                dict["memory"] = value.replace('%','').replace("'","")
                        a.append(dict)
                    a = a[:-1]
                    payload = {}
                    payload['c8y_Docker'] = a
                    self.logger.debug('The following Docker stats where found: %s'% (str(payload)))
                    return json.dumps(payload)
                except Exception as e:
                    self.logger.error('The following error occured: %s'% (str(e)))
            else:
                return None